{-# LANGUAGE GADTs #-}
module Types.Dependency
  ( Action(..)
  , F(..)
  , Dependency(..)
  , makeDependency
  , unListType
  , unTupleType
  , actionTouchesFilesystem
  , formatDependencyTrees
  ) where

import Types.Token (Token(..))
import Types.Value (Value)
import Types.Functions (IsFunction(), IsFunctionIO(..))

import Type.Reflection (Typeable, TypeRep, typeRep)
import Text.Printf (printf)
import Data.Text (Text)
import qualified Data.Text as T

data Action a b where
  Function :: IsFunction f a b => f -> Action a b
  InlineFunction :: Show b => TypeRep a -> TypeRep b -> F a b -> Action a b
  FunctionIO :: IsFunctionIO f a b => f -> Action a b
  Inject :: Value -> Action () a
  FilterComp :: Show a => TypeRep a -> Action ([a], [Bool]) [a]
  UntupleFst :: (Show a, Show b) => TypeRep a -> TypeRep b -> Action (a, b) a
  UntupleSnd :: (Show a, Show b) => TypeRep a -> TypeRep b -> Action (a, b) b
  UnzipFst :: (Show a, Show b) => TypeRep a -> TypeRep b -> Action [(a, b)] [a]
  UnzipSnd :: (Show a, Show b) => TypeRep a -> TypeRep b -> Action [(a, b)] [b]
  MapComp :: (Typeable a, Show a, Typeable b, Show b) => TypeRep a -> TypeRep b -> [Dependency] -> Token a -> Token b -> Action [a] [b]

deriving instance Show (Action a b)

newtype F a b = F (a -> b)

instance Show (F a b) where
  show = const "<function>"

data Dependency where
  Dependency :: (Typeable a, Show a) => TypeRep a -> Token a -> Action a b -> TypeRep b -> Token b -> Dependency
deriving instance Show Dependency

makeDependency :: (Typeable a, Show a, Typeable b, Show b) => Token a -> Action a b -> Token b -> Dependency
makeDependency a action b = Dependency typeRep a action typeRep b

unListType :: Typeable a => TypeRep [a] -> TypeRep a
unListType _ = typeRep

unTupleType :: (Typeable a, Typeable b) => TypeRep (a, b) -> (TypeRep a, TypeRep b)
unTupleType _ = (typeRep, typeRep)

actionTouchesFilesystem :: Action a b -> Bool
actionTouchesFilesystem = \case
  Function _ -> False
  InlineFunction _ _ _ -> False
  FunctionIO f -> functionIOTouchesFilesystem f
  Inject _ -> False
  FilterComp _ -> False
  UntupleFst _ _ -> False
  UntupleSnd _ _ -> False
  UnzipFst _ _ -> False
  UnzipSnd _ _ -> False
  MapComp _ _ subDeps _ _ -> any dependencyTouchesFilesystem subDeps
  where dependencyTouchesFilesystem (Dependency _ _ action _ _) =
          actionTouchesFilesystem action

formatDependencyTrees :: [Dependency] -> Text
formatDependencyTrees = T.concat . (formatDependencyTrees' "")
  where formatDependencyTrees' indentation = concatMap (formatDependencyTree indentation)

        formatDependencyTree indentation (Dependency _ a action _ b) =
          concat [ [ indentation ]
                 , formatToken a
                 , [ " -> " ]
                 , formatToken b
                 , [ ": " ]
                 , formatAction indentation action
                 ]

        formatToken :: Token a -> [Text]
        formatToken = \case
          Token i ->
            [ T.pack (printf "%03d" i) ]
          TupleToken a b ->
            concat [ [ "tup(" ]
                   , formatToken a
                   , [ ", " ]
                   , formatToken b
                   , [ ")" ]
                   ]
          ZipToken a b ->
            concat [ [ "zip(" ]
                   , formatToken a
                   , [ ", " ]
                   , formatToken b
                   , [ ")" ]
                   ]
          ListToken ts ->
            [ "["
            , T.intercalate ", " (map (T.concat . formatToken) ts)
            , "]"
            ]
          NoToken ->
            [ "--" ]

        formatAction :: Text -> Action a b -> [Text]
        formatAction indentation = \case
          MapComp _ _ subDeps innerInput innerOutput ->
            concat [ [ "MapComp(" ]
                   , formatToken innerInput
                   , [ " -> " ]
                   , formatToken innerOutput
                   , [ "):\n" ]
                   , formatDependencyTrees' (T.append indentation "|  ") subDeps
                   ]
          action ->
            [ T.pack (show action)
            , "\n"
            ]
